#!/usr/bin/env python3

import sys
import time
import atexit
import logging
import argparse
import xml.etree.ElementTree as ET

from atex.provisioner.testingfarm import api


# Reuse urllib3 PoolManager configured for heavy Retry attempts
# (because of TestingFarm API reboots, and other transient issues)
http = api._http


def parse_args():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(description="Submit TMT test to Testing Farm")
    parser.add_argument("--repo-url", required=True, help="GitHub repository URL")
    parser.add_argument("--pr-number", required=True, help="Pull request number")
    parser.add_argument("--plan-name", default="/atex_results_plan", help="TMT plan name to run")
    parser.add_argument("--os", default=None, help="OS to test on (e.g., rhel-9)")
    parser.add_argument("--arch", default="x86_64", help="Architecture to test on")
    return parser.parse_args()


def setup_logging():
    """Setup logging configuration."""
    logging.basicConfig(
        level=logging.INFO,  # use DEBUG to see HTTP queries
        stream=sys.stderr,
        format="%(asctime)s %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def build_request_json(args):
    """Build the Testing Farm API request JSON payload."""
    return {
        "test": {
            "fmf": {
                "url": args.repo_url,
                "ref": f"PR{args.pr_number}",
                "name": args.plan_name,
            },
        },
        "environments": [{"arch": args.arch, "os": args.os}],
    }


def get_html_link(artifacts_url):
    """
    Get the HTML link for test results from Testing Farm artifacts.

    Args:
        artifacts_url: URL to Testing Farm artifacts

    Returns:
        str: URL to the HTML results viewer

    Raises:
        RuntimeError: If results.xml or workdir cannot be retrieved
    """
    # Get results.xml for those artifacts, which is a XML representation of the
    # HTML artifacts view and contains links to logs and workdir
    reply = http.request("GET", f"{artifacts_url}/results.xml")
    if reply.status != 200:
        raise RuntimeError("could not get results.xml")

    # Find which log is the workdir and get its URL
    results_xml = ET.fromstring(reply.data)
    for log in results_xml.find("testsuite").find("logs"):
        if log.get("name") == "workdir":
            workdir_url = log.get("href")
            break
    else:
        raise RuntimeError("could not find workdir")

    # TODO: a more reliable way would be to read
    #         {workdir_url}/testing-farm/sanity/execute/results.yaml
    #       as YAML and look for the test name and get its 'data-path'
    #       relative to the /execute/ dir
    return f"{workdir_url}/atex_results_plan/execute/data/guest/default-0/atex_results_test-1/data/index.html?q=status%20IN%20%28%27fail%27%2C%20%27error%27%2C%20%27infra%27%29%20OR%20subtest%20IS%20NULL"


def main():
    """Main function to submit test results to Testing Farm."""
    args = parse_args()
    setup_logging()

    request_json = build_request_json(args)

    # Do faster queries than the default 30 secs, because we don't track
    # many dozens of requests, just one
    class FastRequest(api.Request):
        """
        A request class that executes queries faster than the default 30 seconds.

        This optimization is implemented because the system does not track
        many dozens of requests, typically just one, eliminating the need
        for the standard, longer default timeout.
        """
        api_query_limit = 5

    req = FastRequest()
    req.submit(request_json)
    atexit.register(req.cancel)  # just in case we traceback

    req.wait_for_state("running")

    # Artifacts URL doesn't appear instantly, wait for it
    while "run" not in req:
        time.sleep(FastRequest.api_query_limit)
    while "artifacts" not in req["run"]:
        time.sleep(FastRequest.api_query_limit)

    artifacts_url = req["run"]["artifacts"]
    logging.info(f"artifacts: {artifacts_url}")

    # results.xml appears only after completion
    req.wait_for_state("complete")
    atexit.unregister(req.cancel)

    # Get and print HTML link
    html_link = get_html_link(artifacts_url)
    logging.info(f"HTML: {html_link}")


if __name__ == "__main__":
    main()



from ssg.utils import parse_template_boolean_value


def preprocess(data, lang):
    if data.get("value") is not None and data.get("xccdf_variable") is not None:
        errmsg = (
            f"The template definition of {data['_rule_id']} specifies both "
            f"value and xccdf_variable. This is forbidden."
        )
        raise ValueError(errmsg)
    data["missing_parameter_pass"] = parse_template_boolean_value(
        data, parameter="missing_parameter_pass", default_value=False)
    if "variable_datatype" not in data:
        data["variable_datatype"] = "string"
    return set_variables_for_test_scenarios(data)


def set_variables_for_test_scenarios(data):
    # if no correct value is specified, we will create one for testing purposes
    if not data.get("test_correct_value"):
        if not data.get("value"):
            # this implies XCCDF variable is used
            data["test_correct_value"] = "test_correct_value"
        else:
            data["test_correct_value"] = str(data["value"])
    # if no wrong value is provided, we will create one for testing purposes
    if not data.get("test_wrong_value"):
        data["test_wrong_value"] = "test_wrong_value"
    return data



import ssg.utils
import os


def preprocess(data, lang):
    path = data["path"]
    name = ssg.utils.escape_id(os.path.basename(os.path.normpath(path)))
    data["name"] = name
    if lang == "oval":
        data["path_escaped"] = path.replace("/", "\\/")
    if "key" not in data:
        data["key"] = data["_rule_id"]
    if data["path"].endswith("/"):
        data["filter_type"] = "dir"
    else:
        data["filter_type"] = "path"
    return data



"""
Common functions for processing Requirements Specs in SSG
"""
import re
from typing import Tuple, List

from ssg import utils


class RequirementParser:
    """
    A simple parser for package requirements with version specifiers.
    Handles formats like: package[extra]>=1.0,<2.0
    """

    def __init__(self, target_v: str):
        self.target = target_v
        # First, extract package name and extras
        base_match = re.match(
            r'^(?P<name>[a-zA-Z0-9\-_.]+)(?:\[(?P<extra>[a-zA-Z0-9\-_]+)])?(?P<specs>.*)$',
            target_v
        )

        if not base_match:
            raise ValueError(f"Invalid requirement format: {target_v}")

        self.name = base_match.group('name')
        self.extra = base_match.group('extra')
        specs_str = base_match.group('specs')

        # Parse comma-separated version specifiers
        self.specs_list = []
        if specs_str and specs_str.strip():
            for spec in specs_str.split(','):
                spec = spec.strip()
                if spec:
                    spec_match = re.match(r'^(?P<op>[><!~=]+)\s*(?P<ver>.+)$', spec)
                    if spec_match:
                        self.specs_list.append((spec_match.group('op'), spec_match.group('ver')))
                    else:
                        raise ValueError(f"Invalid version specifier: {spec}")

    def __str__(self):
        return self.target

    @property
    def specs(self) -> List[Tuple[str, str]]:
        return self.specs_list

    @property
    def project_name(self) -> str:
        return self.name

    @property
    def extras(self) -> List[str]:
        if self.extra:
            return [self.extra.lower()]
        return []


def parse_version_into_evr(version):
    """
    Parses a version string into its epoch, version, and release components.

    Args:
        version (str): The version string to parse. It should be in the format
                       'epoch:version-release', where 'epoch' and 'release' are optional.

    Returns:
        dict: A dictionary with keys 'epoch', 'version', and 'release', containing the respective
              parts of the version string. If 'epoch' or 'release' are not present in the input
              string, their values will be None.

    Raises:
        ValueError: If the version string does not match the expected format.
    """
    evr = {"epoch": None, "version": None, "release": None}
    match_version = re.match(r'^(?:(\d+):)?(\d[\d\.]*)(?:-(\d*))?$', version)
    if not match_version:
        raise ValueError("Invalid version specifier {0}".format(version))
    evr["epoch"] = match_version.groups()[0]
    evr["version"] = match_version.groups()[1]
    evr["release"] = match_version.groups()[2]
    return evr


def _spec_to_version_specifier(spec):
    """
    Convert a specification tuple into a VersionSpecifier object.

    Args:
        spec (tuple): A tuple containing an operator and a version string.
                      For example, ('>=', '1.0.0').

    Returns:
        VersionSpecifier: An object representing the version specifier.
    """
    op, ver = spec
    evr = parse_version_into_evr(ver)
    return utils.VersionSpecifier(op, evr)


class Requirement:
    """
    A class to represent a package requirement with version specifications.

    Attributes:
        _req (RequirementParser): The parsed requirement object.
        _specs (utils.VersionSpecifierSet): The set of version specifiers for the requirement.
    """
    def __init__(self, obj: str):
        self._req = RequirementParser(obj)
        self._specs = utils.VersionSpecifierSet(
            [_spec_to_version_specifier(spec) for spec in self._req.specs]
        )

    def __contains__(self, item):
        """Check if a version string satisfies the requirement specs."""
        if not self.has_version_specs():
            return False
        return item in self._specs

    def __str__(self):
        return str(self._req)

    def has_version_specs(self):
        """
        Check if the requirement has version specifications.

        Returns:
            bool: True if there are version specifications, False otherwise.
        """
        return bool(self._req.specs)

    @property
    def ver_specs(self):
        """
        Retrieve the specifications.

        Returns:
            dict: The specifications stored in the instance.
        """
        return self._specs

    @property
    def name(self):
        """
        Retrieve the project name from the requirement specifications.

        Returns:
            str: The name of the project.
        """
        return self._req.project_name

    @property
    def arg(self):
        """
        Retrieve the first extra requirement if available.

        Returns:
            The first element in the extras list if it exists, otherwise None.
        """
        return self._req.extras[0] if self._req.extras else None

    @staticmethod
    def is_parametrized(name):
        """
        Check if a package requirement is parametrized.

        A parametrized package requirement includes extras, which are additional features or
        dependencies that can be optionally included.

        Args:
            name (str): The name of the package requirement to check.

        Returns:
            bool: True if the package requirement is parametrized (includes extras),
                  False otherwise.
        """
        return bool(RequirementParser(name).extras)

    @staticmethod
    def get_base_for_parametrized(name):
        """
        Extracts the base project name from a given parameterized package name.

        Args:
            name (str): The parameterized package name to parse.

        Returns:
            str: The base project name of the package.
        """
        return RequirementParser(name).project_name



"""
Common functions for processing Ansible in SSG
"""

from __future__ import absolute_import

import collections
import copy
import re

from .constants import ansible_version_requirement_pre_task_name
from .constants import min_ansible_version
from . import yaml


def add_minimum_version(ansible_src):
    """
    Adds a minimum Ansible version requirement to an Ansible script.

    This function inserts a pre_task into the provided Ansible script to assert
    that the Ansible version is greater than or equal to a specified minimum version.
    If the script already contains a pre_task or the version check, it will return
    the original script. If a pre_task exists but does not contain the version check,
    it raises a ValueError.

    Args:
        ansible_src (str): The source code of the Ansible script.

    Returns:
        str: The modified Ansible script with the minimum version requirement added.

    Raises:
        ValueError: If a pre_task already exists in the Ansible script but does not
                    contain the version check.
    """
    pre_task = (""" - hosts: all
   pre_tasks:
     - name: %s
       assert:
         that: "ansible_version.full is version_compare('%s', '>=')"
         msg: >
           "You must update Ansible to at least version %s to use this role."
          """ % (ansible_version_requirement_pre_task_name,
                 min_ansible_version, min_ansible_version))

    if ' - hosts: all' not in ansible_src:
        return ansible_src

    if 'pre_task' in ansible_src:
        if 'ansible_version.full is version_compare' in ansible_src:
            return ansible_src

        raise ValueError(
            "A pre_task already exists in ansible_src; failing to process: %s" %
            ansible_src)

    return ansible_src.replace(" - hosts: all", pre_task, 1)


def remove_too_many_blank_lines(ansible_src):
    """
    Condenses three or more consecutive empty lines into two empty lines.

    Args:
        ansible_src (str): The source string from an Ansible file.

    Returns:
        str: The modified string with excessive blank lines reduced.
    """
    return re.sub(r'\n{4,}', '\n\n\n', ansible_src, count=0, flags=re.MULTILINE)


def remove_trailing_whitespace(ansible_src):
    """
    Remove trailing whitespace from each line in the given Ansible source string.

    Args:
        ansible_src (str): The Ansible source code as a string.

    Returns:
        str: The Ansible source code with trailing whitespace removed from each line.
    """

    return re.sub(r'[ \t]+$', '', ansible_src, count=0, flags=re.MULTILINE)


package_facts_task = collections.OrderedDict([
    ('name', 'Gather the package facts'),
    ('ansible.builtin.package_facts', {'manager': 'auto'}),
    ('tags', ['always'])
])
service_facts_task = collections.OrderedDict([
    ('name', 'Gather the service facts'),
    ('ansible.builtin.service_facts', None),
    ('tags', ['always'])
])


def task_is(task, names):
    """
    Check if the task is one of the given names.

    Args:
        task (dict): The task to check.
        names (list): List of task names to check against.

    Returns:
        bool: True if the task contains any of the given names as keys, False otherwise.
    """
    for name in names:
        if name in task:
            return True
    return False


class AnsibleSnippetsProcessor:
    """
    Processes Ansible snippets to optimize package management tasks.

    This class processes a collection of Ansible snippets by:
    - Collecting package-related tasks for batch processing
    - Skipping redundant package_facts tasks
    - Handling block tasks recursively
    - Preserving other tasks as-is
    """

    def __init__(self, all_snippets):
        """
        Initialize the processor with a collection of snippets.

        Args:
            all_snippets (list): List of Ansible snippet strings to process.
        """
        self.all_snippets = all_snippets
        self.package_tasks = []
        self.service_tasks = []
        self.other_tasks = []

    def _process_task(self, task):
        """
        Process a single task, determining how to handle it.

        Args:
            task (dict): The task to process.

        Returns:
            dict or None: The processed task, or None if the task should be skipped.
        """

        if task_is(task, ["block"]):
            # Process block tasks recursively
            new_block = []
            for subtask in task["block"]:
                if self._process_task(subtask) is not None:
                    new_block.append(subtask)
            task["block"] = new_block
            if "special_service_block" in task.get("tags", []):
                # Collect service tasks to be processed later
                self.service_tasks.append(task)
                return None
            return task if new_block else None
        if task_is(task, ["ansible.builtin.package_facts", "package_facts"]):
            # Skip package_facts tasks because they will be replaced by
            # a single package_facts task that will be added later
            return None
        if task_is(task, ["ansible.builtin.service_facts", "service_facts"]):
            # Skip service_facts tasks because they will be replaced by
            # a single service_facts task that will be added later
            return None
        if task_is(task, ["ansible.builtin.package", "package"]):
            # Collect package tasks to be processed later
            self.package_tasks.append(task)
            return None
        return task

    def _process_snippet(self, snippet):
        """
        Process a single snippet, extracting tasks from it.

        Args:
            snippet (str): The YAML snippet string to process.
        """
        tasks = yaml.ordered_load(snippet)
        for task in tasks:
            if self._process_task(task) is not None:
                self.other_tasks.append(task)

    def process_snippets(self):
        """
        Process all snippets provided during initialization.
        """
        for snippet in self.all_snippets:
            self._process_snippet(snippet)

    def get_ansible_tasks(self):
        """
        Get the final list of processed Ansible tasks.

        Package facts tasks are added at the beginning and end of package tasks,
        then combined with other tasks.

        Returns:
            list: Combined list of all processed tasks.
        """
        return [package_facts_task, *self.package_tasks, copy.deepcopy(package_facts_task), *self.service_tasks, service_facts_task, *self.other_tasks]



#!/usr/bin/env python3

import argparse
import json
import logging
import os
import pathlib
import sys
from typing import TypeVar, Generator, Set, Dict
import multiprocessing

import ssg.constants
import ssg.environment
import ssg.jinja
import ssg.utils
import ssg.yaml
import ssg.templates

SSG_ROOT = str(pathlib.Path(__file__).resolve().parent.parent.absolute())
JOB_COUNT = multiprocessing.cpu_count()
T = TypeVar("T")
TESTS_CONFIG_NAME = "test_config.yml"


def _create_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Converts built content tests to be rendered.")
    parser.add_argument("--build-config-yaml", required=True, type=str,
                        help="YAML file with information about the build configuration. "
                             "e.g.: ~/scap-security-guide/build/build_config.yml")
    parser.add_argument("--product-yaml", required=True, type=str,
                        help="YAML file with information about the product we are building. "
                             "e.g.: ~/scap-security-guide/build/rhel10/product.yml")
    parser.add_argument("--output", required=True, type=str,
                        help="Output path"
                             "e.g.:  ~/scap-security-guide/build/rhel10/tests")
    parser.add_argument("--resolved-rules-dir", required=True, type=str,
                        help="Directory with <rule-id>.yml resolved rule YAMLs "
                             "e.g.: ~/scap-security-guide/build/rhel10/rules")
    parser.add_argument("--log-level", action="store", type=str, default="ERROR",
                        choices=["ERROR", "WARNING", "INFO", "DEBUG", "TRACE"],
                        help="What level to log at. Defaults to ERROR.")
    parser.add_argument("--root", default=SSG_ROOT,
                        help=f"Path to the project. Defaults to {SSG_ROOT}")
    parser.add_argument("--jobs", "-j", type=int, default=JOB_COUNT,
                        help=f"Number of cores to use. Defaults to {JOB_COUNT} on this system.")
    return parser


def _write_path(file_contents: str, output_path: os.PathLike) -> None:
    with open(output_path, "w") as file:
        file.write(file_contents.replace('&amp;', '&'))
        file.write("\n")


def _is_test_file(filename: str) -> bool:
    return filename.endswith(('.pass.sh', '.fail.sh', '.notapplicable.sh'))


def _get_deny_templated_scenarios(test_config_path: pathlib.Path) -> Set[str]:
    if test_config_path.exists():
        test_config = ssg.yaml.open_raw(str(test_config_path.absolute()))
        deny_templated_scenarios = test_config.get('deny_templated_scenarios', set())
        return deny_templated_scenarios
    return set()


def _process_shared_file(env_yaml: dict, file: pathlib.Path, shared_output_path: pathlib.Path) \
        -> None:
    file_contents = ssg.jinja.process_file(str(file.absolute()), env_yaml)
    shared_script_path = shared_output_path / file.name
    _write_path(file_contents, shared_script_path)


def _copy_and_process_shared(env_yaml: dict, output_path: pathlib.Path, root_path: pathlib.Path) \
        -> None:
    tests_shared_root = root_path / "tests" / "shared"
    shared_output_path = output_path / "shared"
    shared_output_path.mkdir(parents=True, exist_ok=True)
    for file in tests_shared_root.iterdir():  # type: pathlib.Path
        # We only support one level deep, this avoids recursive functions
        if file.is_dir():
            for sub_file in file.iterdir():
                shared_output_path_sub = shared_output_path / file.name
                shared_output_path_sub.mkdir(parents=True, exist_ok=True)
                _process_shared_file(env_yaml, sub_file, shared_output_path_sub)
        else:
            _process_shared_file(env_yaml, file, shared_output_path)


def _get_platform_from_file_contents(file_contents: str) -> str:
    # Some tests don't have an explict platform assume always applicable
    platform = "multi_platform_all"
    for line in file_contents.split("\n"):
        if line.startswith('# platform'):
            platform_parts = line.split('=')
            if len(platform_parts) == 2:
                platform = platform_parts[1]
                break
    return platform.strip()


def _process_local_tests(product: str, env_yaml: dict, rule_output_path: pathlib.Path,
                         rule_tests_root: pathlib.Path) -> None:
    logger = logging.getLogger()
    for test in rule_tests_root.iterdir():  # type: pathlib.Path
        if test.is_dir():
            logger.warning("Skipping directory %s in rule %s", test.name,
                           rule_output_path.name)
            continue
        if not _is_test_file(test.name):
            file_contents = ssg.jinja.process_file(str(test.absolute()),
                                                   env_yaml)
            output_file = rule_output_path / test.name
            rule_output_path.mkdir(parents=True, exist_ok=True)
            _write_path(file_contents, output_file)
        file_contents = test.read_text()
        platform = _get_platform_from_file_contents(file_contents)
        if ssg.utils.is_applicable_for_product(platform, product):
            content = ssg.jinja.process_file(str(test.absolute()), env_yaml)
            rule_output_path.mkdir(parents=True, exist_ok=True)
            _write_path(content, rule_output_path / test.name)


def _get_test_dir_config(rule_path: pathlib.Path) -> Dict:
    test_config = dict()
    rule_root = rule_path.parent
    tests_dir = rule_root / "tests"
    test_config_path = tests_dir / TESTS_CONFIG_NAME
    if test_config_path.exists():
        test_config = ssg.yaml.open_raw(test_config_path)
    return test_config


def _process_templated_tests(env_yaml: Dict, rendered_rule_obj: Dict, templates_root: pathlib.Path,
                             rule_output_path: pathlib.Path):
    logger = logging.getLogger()
    rule_path = pathlib.Path(rendered_rule_obj['definition_location'])
    product = rule_output_path.parent.parent.name
    rule_id = rule_path.parent.name
    if "name" not in rendered_rule_obj["template"]:
        raise ValueError(f"Invalid template config on rule {rule_id}")
    template_name = rendered_rule_obj["template"]["name"]
    template_root = templates_root / template_name
    template_tests_root = template_root / "tests"

    if not template_tests_root.exists():
        logger.debug("Template %s doesn't have tests. Skipping for rule %s.",
                     template_name, rule_id)
        return
    test_config = _get_test_dir_config(rule_path)
    all_templated_tests = set(x.name for x in template_tests_root.iterdir())
    templated_tests = ssg.utils.select_templated_tests(test_config, all_templated_tests)
    for test_name in templated_tests:  # type: str
        test = template_tests_root / test_name
        if not test.name.endswith(".sh"):
            logger.warning("Skipping %s for %s as it isn't a test scenario",
                           test.name, rule_id)
            continue
        template = ssg.templates.Template.load_template(str(templates_root.absolute()),
                                                        template_name)
        rendered_rule_obj["template"]["vars"]["_rule_id"] = rule_id
        template_parameters = template.preprocess(rendered_rule_obj["template"]["vars"], "test")
        env_yaml = env_yaml.copy()
        jinja_dict = ssg.utils.merge_dicts(env_yaml, template_parameters)
        file_contents = ssg.jinja.process_file(str(test.absolute()), jinja_dict)
        platform = _get_platform_from_file_contents(file_contents)
        if ssg.utils.is_applicable_for_product(platform, product):
            rule_output_path.mkdir(parents=True, exist_ok=True)
            test_output_path = rule_output_path / test.name
            _write_path(file_contents, test_output_path)
            logger.debug("Wrote scenario %s for rule %s", test.name, rule_id)
        else:
            logger.warning("Skipping scenario %s for rule %s as it not applicable to %s",
                           test.name, rule_id, product)


def _process_rules(env_yaml: Dict, output_path: pathlib.Path,
                   templates_root: pathlib.Path, product_rules: list,
                   resolved_root: pathlib.Path) -> None:
    product = resolved_root.parent.name
    for rule_id in product_rules:
        rule_file = resolved_root / f'{rule_id}.json'

        with open(rule_file, "r") as file:
            rendered_rule_obj = json.load(file)
        rule_path = pathlib.Path(rendered_rule_obj["definition_location"])
        rule_root = rule_path.parent

        rule_tests_root = rule_root / "tests"
        rule_output_path = output_path / rule_id
        if rendered_rule_obj["template"] is not None:
            _process_templated_tests(env_yaml, rendered_rule_obj, templates_root, rule_output_path)
        if rule_tests_root.exists():
            _process_local_tests(product, env_yaml, rule_output_path, rule_tests_root)


def _get_rules_in_profile(built_profiles_root) -> Generator[str, None, None]:
    for profile_file in built_profiles_root.iterdir():  # type: pathlib.Path
        if not profile_file.name.endswith(".profile"):
            continue
        with open(str(profile_file.absolute()), "r") as file:
            profile_data = json.load(file)
        for selection in profile_data["selections"]:
            if "=" not in selection:
                yield selection


def main() -> int:
    args = _create_arg_parser().parse_args()
    logging.basicConfig(level=logging.getLevelName(args.log_level))
    env_yaml = ssg.environment.open_environment(args.build_config_yaml, args.product_yaml)
    ssg.jinja.initialize(env_yaml)

    root_path = pathlib.Path(args.root).resolve()
    output_path = pathlib.Path(args.output).resolve()
    resolved_rules_dir = pathlib.Path(args.resolved_rules_dir)
    if not resolved_rules_dir.exists() or not resolved_rules_dir.is_dir():
        logging.error("Unable to find product at %s", str(resolved_rules_dir))
        logging.error("Is the product built?")
        return 1

    output_path.mkdir(parents=True, exist_ok=True)
    _copy_and_process_shared(env_yaml, output_path, root_path)

    built_profiles_root = resolved_rules_dir.parent / "profiles"
    rules_in_profiles = list(set(_get_rules_in_profile(built_profiles_root)))

    templates_root = root_path / "shared" / "templates"
    processes = list()
    for chunk in range(args.jobs):
        process_args = (env_yaml, output_path, templates_root,
                        rules_in_profiles[chunk::args.jobs], resolved_rules_dir)
        process = multiprocessing.Process(target=_process_rules, args=process_args)
        processes.append(process)
        process.start()
    for process in processes:
        process.join()
    # Write a file for CMake
    # So we don't have a dependency on a folder
    done_file = output_path / ".test_done"
    done_file.touch()
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import argparse
import pathlib
import sys
import xml.etree.ElementTree as ET
from ssg.constants import XCCDF12_NS
from typing import Generator


SSG_ROOT = pathlib.Path(__file__).resolve().parent.parent.absolute()
BUILT_ROOT = SSG_ROOT.joinpath('build')

NS = {'xccdf-1.2': XCCDF12_NS}


def _create_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('--old-release-dir', required=True,
                        help='Path to built content from last release')
    parser.add_argument('--build-root', default=str(BUILT_ROOT),
                        help=f'Path to the current build root. Defaults to {BUILT_ROOT}')
    parser.add_argument('--product', help='Product to check')
    return parser


def _get_rules(root: ET.Element) -> Generator[str, None, None]:
    for rule in root.findall('.//xccdf-1.2:Rule', namespaces=NS):
        yield rule.get('id', '')


def main() -> int:
    args = _create_arg_parser().parse_args()
    old_build_root = pathlib.Path(args.old_release_dir)
    old_rules_datastream_path = old_build_root / f'ssg-{args.product}-ds.xml'
    old_ds_root = ET.parse(old_rules_datastream_path).getroot()
    old_rules = set(_get_rules(old_ds_root))

    new_build_root = pathlib.Path(args.build_root)
    new_rules_datastream_path = new_build_root / f'ssg-{args.product}-ds.xml'
    new_ds_root = ET.parse(new_rules_datastream_path).getroot()
    new_rules = set(_get_rules(new_ds_root))

    missing = old_rules - new_rules
    if len(missing) != 0:
        for rule in missing:
            print(f'{rule} is missing in new data stream', file=sys.stderr)
        return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



import os
import re
import textwrap
from collections import OrderedDict

import pytest

import ssg.ansible
from ssg.constants import min_ansible_version


def strings_equal_except_whitespaces(left, right):
    return re.sub(r'\s', '', left) == re.sub(r'\s', '', right)


def test_add_minimum_version():
    good_snippet = """
    # a comment
     - hosts: all
       vars:
       tasks:
    """
    good_snippet = textwrap.dedent(good_snippet)

    bad_snippet = """
    # a comment
     - hosts: all
       pre_tasks:
         - name: something_else
           assert:
             that: "2 > 3"
             msg: "two is less than three!"
       vars:
       tasks:
    """
    bad_snippet = textwrap.dedent(bad_snippet)

    unknown_snippet = """
    I don't think this is YAML
    """

    processed_snippet = """
    # a comment
     - hosts: all
       pre_tasks:
         - name: Verify Ansible meets SCAP-Security-Guide version requirements.
           assert:
             that: "ansible_version.full is version_compare('{min_version}', '>=')"
             msg: >
               "You must update Ansible to at least version {min_version} to use this role."

       vars:
       tasks:
    """.format(min_version=min_ansible_version)
    processed_snippet = textwrap.dedent(processed_snippet)

    output = ssg.ansible.add_minimum_version(good_snippet)
    assert strings_equal_except_whitespaces(output, processed_snippet)

    with pytest.raises(ValueError):
        ssg.ansible.add_minimum_version(bad_snippet)

    assert ssg.ansible.add_minimum_version(unknown_snippet) == unknown_snippet

    assert ssg.ansible.add_minimum_version(processed_snippet) == processed_snippet


def test_task_is():
    # Test with task containing one of the names
    task_with_package = {"ansible.builtin.package": {"name": "vim"}}
    assert ssg.ansible.task_is(task_with_package, ["package", "ansible.builtin.package"])

    # Test with task containing different name
    assert ssg.ansible.task_is(task_with_package, ["ansible.builtin.package"])

    # Test with task not containing any of the names
    task_with_copy = {"copy": {"src": "file", "dest": "/tmp"}}
    assert not ssg.ansible.task_is(task_with_copy, ["package", "ansible.builtin.package"])

    # Test with empty names list
    assert not ssg.ansible.task_is(task_with_package, [])

    # Test with empty task
    assert not ssg.ansible.task_is({}, ["package"])


class TestAnsibleSnippetsProcessor:

    def test_initialization(self):
        """Test processor initialization."""
        snippets = ["snippet1", "snippet2"]
        processor = ssg.ansible.AnsibleSnippetsProcessor(snippets)

        assert processor.all_snippets == snippets
        assert processor.package_tasks == []
        assert processor.other_tasks == []

    def test_process_task_package(self):
        """Test processing of package tasks."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        package_task = {"ansible.builtin.package": {"name": "vim"}}
        result = processor._process_task(package_task)

        assert result is None  # Package tasks should be skipped from immediate processing
        assert package_task in processor.package_tasks

    def test_process_task_package_facts(self):
        """Test processing of package_facts tasks."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        package_facts_task = {"ansible.builtin.package_facts": {"manager": "auto"}}
        result = processor._process_task(package_facts_task)

        assert result is None  # package_facts tasks should be skipped
        assert len(processor.package_tasks) == 0  # Should not be added to package_tasks

    def test_process_task_package_facts_short_name(self):
        """Test processing of package_facts tasks with short name."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        package_facts_task = {"package_facts": {"manager": "auto"}}
        result = processor._process_task(package_facts_task)

        assert result is None  # package_facts tasks should be skipped

    def test_process_task_service_facts(self):
        """Test processing of service_facts tasks."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        service_facts_task = {"ansible.builtin.service_facts": None}
        result = processor._process_task(service_facts_task)

        assert result is None  # service_facts tasks should be skipped
        assert len(processor.other_tasks) == 0  # Should not be added to other_tasks

    def test_process_task_other(self):
        """Test processing of other tasks."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        copy_task = {"copy": {"src": "file", "dest": "/tmp"}}
        result = processor._process_task(copy_task)

        assert result == copy_task  # Other tasks should be returned as-is
        assert len(processor.package_tasks) == 0
        assert len(processor.other_tasks) == 0  # _process_task doesn't add to other_tasks

    def test_process_task_block_empty(self):
        """Test processing of empty block tasks."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        block_task = {"block": []}
        result = processor._process_task(block_task)

        assert result is None  # Empty blocks should be filtered out

    def test_process_task_block_with_content(self):
        """Test processing of block tasks with content."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        block_task = {
            "block": [
                {"copy": {"src": "file", "dest": "/tmp"}},
                {"ansible.builtin.package": {"name": "vim"}},
                {"ansible.builtin.package_facts": {"manager": "auto"}}
            ]
        }
        result = processor._process_task(block_task)

        # Should return block with only the copy task (package tasks are collected, package_facts skipped)
        assert result is not None
        assert "block" in result
        assert len(result["block"]) == 1
        assert result["block"][0] == {"copy": {"src": "file", "dest": "/tmp"}}
        assert len(processor.package_tasks) == 1

    def test_process_task_block_all_filtered(self):
        """Test processing of block tasks where all subtasks are filtered."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        block_task = {
            "block": [
                {"ansible.builtin.package": {"name": "vim"}},
                {"ansible.builtin.package_facts": {"manager": "auto"}}
            ]
        }
        result = processor._process_task(block_task)

        assert result is None  # Block should be None if all subtasks are filtered
        assert len(processor.package_tasks) == 1

    def test_process_snippet(self):
        """Test processing of a complete snippet."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])

        snippet = """
        - name: Install vim
          ansible.builtin.package:
            name: vim
        - name: Gather facts
          ansible.builtin.package_facts:
            manager: auto
        - name: Copy file
          copy:
            src: file
            dest: /tmp
        """

        processor._process_snippet(snippet)

        assert len(processor.package_tasks) == 1
        assert len(processor.other_tasks) == 1
        assert processor.other_tasks[0]["copy"]["src"] == "file"

    def test_process_snippets(self):
        """Test processing multiple snippets."""
        snippet1 = """
        - name: Install vim
          ansible.builtin.package:
            name: vim
        """

        snippet2 = """
        - name: Copy file
          copy:
            src: file
            dest: /tmp
        """

        processor = ssg.ansible.AnsibleSnippetsProcessor([snippet1, snippet2])
        processor.process_snippets()

        assert len(processor.package_tasks) == 1
        assert len(processor.other_tasks) == 1

    def test_get_ansible_tasks(self):
        """Test getting the final ansible tasks."""
        snippet = """
        - name: Install vim
          ansible.builtin.package:
            name: vim
        - name: Copy file
          copy:
            src: file
            dest: /tmp
        """

        processor = ssg.ansible.AnsibleSnippetsProcessor([snippet])
        processor.process_snippets()

        tasks = processor.get_ansible_tasks()

        # Should have: package_facts_task + package_task + package_facts_task + service_facts_task + other_task
        assert len(tasks) == 5
        assert tasks[0] == ssg.ansible.package_facts_task  # First package facts task
        assert "ansible.builtin.package" in tasks[1]  # Package task
        assert tasks[2] == ssg.ansible.package_facts_task  # Second package facts task
        assert "ansible.builtin.service_facts" in tasks[3]  # Service facts task
        assert "copy" in tasks[4]  # Other task

    def test_get_ansible_tasks_no_package_tasks(self):
        """Test getting ansible tasks when there are no package tasks."""
        snippet = """
        - name: Copy file
          copy:
            src: file
            dest: /tmp
        """

        processor = ssg.ansible.AnsibleSnippetsProcessor([snippet])
        processor.process_snippets()

        tasks = processor.get_ansible_tasks()

        # Should have: package_facts_task + package_facts_task + service_facts_task + other_task
        assert len(tasks) == 4
        assert tasks[0] == ssg.ansible.package_facts_task
        assert tasks[1] == ssg.ansible.package_facts_task
        assert tasks[2] == ssg.ansible.service_facts_task
        assert "copy" in tasks[3]  # Other task

    def test_empty_snippets(self):
        """Test processing empty snippets list."""
        processor = ssg.ansible.AnsibleSnippetsProcessor([])
        processor.process_snippets()

        tasks = processor.get_ansible_tasks()

        # Should have only the two package facts tasks and service facts task
        assert len(tasks) == 3
        assert tasks[0] == ssg.ansible.package_facts_task
        assert tasks[1] == ssg.ansible.package_facts_task
        assert tasks[2] == ssg.ansible.service_facts_task

    def test_malformed_snippet(self):
        """Test handling of malformed YAML snippets."""
        processor = ssg.ansible.AnsibleSnippetsProcessor(["invalid: yaml: content:"])

        # This should raise SystemExit when trying to parse invalid YAML
        with pytest.raises(SystemExit):
            processor.process_snippets()

    def test_process_task_special_service_block(self):
      """Test processing of special service block tasks."""
      processor = ssg.ansible.AnsibleSnippetsProcessor([])

      service_task = {
        "name": "Enable service sshd",
        "block": [{"name":"Gather package facts", "ansible.builtin.package_facts": {"manager": "auto"}}, {"name": "Enable service sshd", "ansible.builtin.systemd": {"name": "sshd"}}],
        "tags": ["special_service_block"]
      }
      result = processor._process_task(service_task)

      assert result is None  # Service tasks should be skipped
      assert len(processor.service_tasks) == 1
      assert processor.service_tasks[0] == service_task
      assert "ansible.builtin.package_facts" not in processor.service_tasks[0]["block"][0]



def set_variables_for_test_scenarios(data):
    # if no correct value is specified, we will create one for testing purposes
    if not data.get("test_correct_value"):
        if not data.get("value"):
            # this implies XCCDF variable is used
            data["test_correct_value"] = "test_correct_value"
        else:
            data["test_correct_value"] = str(data["value"])
    # if no wrong value is provided, we will create one for testing purposes
    if not data.get("test_wrong_value"):
        data["test_wrong_value"] = "test_wrong_value"
    return data


def preprocess(data, lang):
    if data.get("value") is not None and data.get("xccdf_variable") is not None:
        errmsg = ("The template definition of {0} specifies both value and xccdf_variable."
                  "This is forbidden.".format(data["_rule_id"]))
        raise ValueError(errmsg)
    if "sep" not in data:
        data["sep"] = " = "

    if "prefix_regex" not in data:
        data["prefix_regex"] = r"^\s*"

    if "sep_regex" not in data:
        data["sep_regex"] = r"\s*=\s*"

    if "app" not in data:
        data["app"] = ""

    if "variable_datatype" not in data:
        data["variable_datatype"] = "string"

    data = set_variables_for_test_scenarios(data)
    return data



#!/usr/bin/env python3

import sys
import time
import json
import lzma
import atexit
import signal
import logging
import argparse
import contextlib
from pathlib import Path

from atex.provisioner.testingfarm import TestingFarmProvisioner
from atex.orchestrator.contest import ContestOrchestrator
from atex.aggregator.json import LZMAJSONAggregator
from atex.fmf import FMFTests


def kv_pair(keyval):
    if "=" not in keyval:
        raise argparse.ArgumentTypeError(f"expected KEY=VALUE, got: {keyval}")
    return keyval.split("=", 1)

def parse_args():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(description="Run tests on Testing Farm using atex")
    parser.add_argument("--contest-dir", required=True, help="Path to contest repository")
    parser.add_argument("--content-dir", required=True, help="Path to built content directory")
    parser.add_argument("--plan", required=True, help="TMT plan to run (e.g., daily|ci-gating|weekly)")
    parser.add_argument("--compose", required=True, help="compose (e.g., Centos-Stream-9)")
    parser.add_argument("--arch", default="x86_64", help="Architecture")
    parser.add_argument("--os-major-version", required=True, help="OS Major Version (8|9|10)")
    parser.add_argument("--tests", nargs="*", help="Specific tests to run (optional, runs all if not specified)")
    parser.add_argument("--timeout", type=int, default=120, help="Timeout in minutes")
    parser.add_argument("--max-remotes", type=int, default=10, help="Maximum number of parallel test executions")
    parser.add_argument("--reruns", type=int, default=1, help="Number of test reruns on failure")
    parser.add_argument("--tag", type=kv_pair, default=[], action="append", help="Additional tag(s) to store in TF Request")
    return parser.parse_args()


class OrchestratorLogFilter(logging.Filter):
    """
    Filter (show) only messages from the atex.orchestrator logger,
    show all warnings (and above) from everywhere,
    show all from the root logger (logging.*).
    """
    def filter(self, record):
        return (
            record.levelno >= logging.WARNING
            or record.name.startswith("atex.orchestrator")
            or record.name == "root"
        )


def setup_logging():
    """Setup logging configuration with console and file handlers."""
    # console - keep it brief, just basic orchestration + warnings
    console_log = logging.StreamHandler(sys.stderr)
    console_log.setLevel(logging.INFO)
    console_log.addFilter(OrchestratorLogFilter())
    console_log.setFormatter(logging.Formatter(
        fmt="%(asctime)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    ))

    # debug log - store ALL debugging info, compressed
    debug_log_fobj = lzma.open("atex_debug.log.xz", "wt")
    atexit.register(debug_log_fobj.close)
    file_log = logging.StreamHandler(debug_log_fobj)
    file_log.setLevel(logging.DEBUG)
    file_log.setFormatter(logging.Formatter(
        fmt="%(asctime)s %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    ))

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=(console_log, file_log),
        force=True,
    )


def setup_signal_handlers():
    """Setup signal handlers for graceful abort."""
    def abort_on_signal(signum, _):
        logging.error(f"got signal {signum}, aborting")
        raise SystemExit(1)

    signal.signal(signal.SIGTERM, abort_on_signal)
    signal.signal(signal.SIGHUP, abort_on_signal)


def main():
    """Main function to run tests on Testing Farm."""
    setup_logging()
    setup_signal_handlers()

    args = parse_args()

    # Variables exported to tests
    test_env = {
        "CONTEST_CONTENT": ContestOrchestrator.content_dir_on_remote,
        "CONTEST_VERBOSE": "2",
    }

    with contextlib.ExitStack() as stack:
        # Load FMF tests from contest directory
        fmf_tests = FMFTests(
            args.contest_dir,
            args.plan,
            names=args.tests or None,
            context={
                "distro": f"centos-stream-{args.os_major_version}",
                "arch": args.arch,
            },
        )

        logging.info(f"plan: {args.plan}")
        logging.info(f"os major version: {args.os_major_version}")
        logging.info(f"arch: {args.arch}")
        logging.info(f"compose: {args.compose}")
        logging.info("will run:")
        for test in fmf_tests.tests:
            logging.info(f"    {test}")

        # Setup result aggregator
        output_results = f"results-centos-stream-{args.os_major_version}-{args.arch}.json.xz"
        output_files = f"files-centos-stream-{args.os_major_version}-{args.arch}"
        partial_runs = Path(output_files) / "old_runs"
        aggregator = LZMAJSONAggregator(output_results, output_files)
        stack.enter_context(aggregator)

        partial_runs.mkdir(parents=True, exist_ok=True)

        platform_name = f"cs{args.os_major_version}@{args.arch}"

        # Setup Testing Farm provisioner
        prov = TestingFarmProvisioner(
            compose=args.compose,
            arch=args.arch,
            max_retries=2,
            timeout=args.timeout,
            tags=dict(args.tag),
        )

        # Setup Contest orchestrator
        orchestrator = ContestOrchestrator(
            platform=platform_name,
            fmf_tests=fmf_tests,
            provisioners=[prov],
            aggregator=aggregator,
            tmp_dir=partial_runs,
            max_remotes=args.max_remotes,
            max_spares=2,
            max_reruns=args.reruns,
            content_dir=args.content_dir,
            env=test_env,
        )
        stack.enter_context(orchestrator)

        logging.info("Starting test execution...")
        next_writeout = time.monotonic() + 600
        while orchestrator.serve_once():
            if time.monotonic() > next_writeout:
                logging.info(
                    "STATISTICS: "
                    f"queued: {len(orchestrator.to_run)}/{len(fmf_tests.tests)} tests, "
                    f"running: {len(orchestrator.running_tests)} tests",
                )
                next_writeout = time.monotonic() + 600
            time.sleep(1)

        logging.info("Test execution completed!")

    # Log final output locations
    logging.info(f"Results written to: {output_results}")
    logging.info(f"Test files in: {output_files}")

    # Read back the compressed JSON results and exit with non-0 if anything failed
    with lzma.open(output_results, "rt") as results:
        for line in results:
            fields = json.loads(line)
            # [platform, status, test name, subtest name, files, note]
            if fields[1] in ("fail", "error", "infra"):
                logging.error("failures found in the results, exiting with 1")
                sys.exit(1)


if __name__ == "__main__":
    main()



from ssg.utils import parse_template_boolean_value


def preprocess(data, lang):
    data["missing_parameter_pass"] = parse_template_boolean_value(
        data, parameter="missing_parameter_pass", default_value=False)
    data["no_quotes"] = parse_template_boolean_value(
        data, parameter="no_quotes", default_value=False)
    data["missing_config_file_fail"] = parse_template_boolean_value(
        data, parameter="missing_config_file_fail", default_value=True)
    return data
